from marshmallow import Schema, fields, post_load
from ..resource import Resource, _add_method
from collections import namedtuple


class Ping(Resource):
    """
    https://dev.chartmogul.com/docs/authentication
    """
    _path = "/ping"

    class _Schema(Schema):
        data = fields.String()

        @post_load
        def make(self, data):
            return Ping(**data)

    _schema = _Schema(strict=True)


_add_method(Ping, "ping", "get")



import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry

METHOD_WHITELIST = ['HEAD', 'GET', 'POST', 'PUT', 'DELETE', 'OPTIONS', 'TRACE']
STATUS_FORCELIST = (429, 500, 502, 503, 504, 520, 524)

def requests_retry_session(retries=20, backoff_factor=2, session=None,):
    session = session or requests.Session()
    adapter = _retry_adapter(retries, backoff_factor)
    session.mount('https://', adapter)
    return session

def _retry_adapter(retries, backoff_factor):
    retry = Retry(
        total=retries,
        read=retries,
        connect=retries,
        status=retries,
        method_whitelist=METHOD_WHITELIST,
        status_forcelist=STATUS_FORCELIST,
        backoff_factor=backoff_factor,
    )
    return HTTPAdapter(max_retries=retry)



from marshmallow import Schema, fields, post_load
from ..resource import Resource, DataObject, _add_method
from collections import namedtuple


class Summary(DataObject):
    """
    Optional information about a series of metrics.
    """
    class _Schema(Schema):
        current = fields.Number()
        previous = fields.Number()
        percentage_change = fields.Number(load_from='percentage-change')

        @post_load
        def make(self, data):
            return Summary(**data)

    _schema = _Schema(strict=True)


class Metrics(Resource):
    """
    https://dev.chartmogul.com/v1.0/reference#introduction-metrics-api
    """
    _path = "/metrics/all"
    _root_key = 'entries'
    _many_cls = namedtuple('Metrics', [_root_key, 'summary'])
    _many_cls.__new__.__defaults__ = (None,) * len(_many_cls._fields)

    class _Schema(Schema):
        """
        Fields are optional, so a subset present is good enough
        """
        date = fields.Date()
        customer_churn_rate = fields.Number(load_from='customer-churn-rate')
        mrr_churn_rate = fields.Number(load_from='mrr-churn-rate')
        ltv = fields.Number()
        customers = fields.Number()
        asp = fields.Number()
        arpa = fields.Number()
        arr = fields.Number()
        mrr = fields.Number()
        # MRR only
        mrr_new_business = fields.Number(load_from='mrr-new-business')
        mrr_expansion = fields.Number(load_from='mrr-expansion')
        mrr_contraction = fields.Number(load_from='mrr-contraction')
        mrr_churn = fields.Number(load_from='mrr-churn')
        mrr_reactivation = fields.Number(load_from='mrr-reactivation')

        @post_load
        def make(self, data):
            return Metrics(**data)

    _schema = _Schema(strict=True)

    @classmethod
    def _many(cls, entries, **kwargs):
        if 'summary' in kwargs:
            kwargs['summary'] = Summary._schema.load(kwargs['summary']).data
        return cls._many_cls(entries, **kwargs)


_add_method(Metrics, "mrr", "get", path='/metrics/mrr')
_add_method(Metrics, "arr", "get", path='/metrics/arr')
_add_method(Metrics, "arpa", "get", path='/metrics/arpa')
_add_method(Metrics, "asp", "get", path='/metrics/asp')
_add_method(Metrics, "customer_count", "get", path='/metrics/customer-count')
_add_method(Metrics, "customer_churn_rate", "get", path='/metrics/customer-churn-rate')
_add_method(Metrics, "mrr_churn_rate", "get", path='/metrics/mrr-churn-rate')
_add_method(Metrics, "ltv", "get", path='/metrics/ltv')



"""
Beside the Customer class this module uses also subdocument classes.
These are used only on deserializing the server response, so the client has
proper types (eg. parsed dates instead of strings). Creating customer only
requires Customer class and simple dict/array data.
Validation is done on server.
"""
from marshmallow import Schema, fields, post_load
from ..resource import Resource, DataObject, _add_method
from collections import namedtuple
from .attributes import Attributes


class Address(DataObject):

    class _Schema(Schema):
        address_zip = fields.String(allow_none=True)
        city = fields.String(allow_none=True)
        country = fields.String(allow_none=True)
        state = fields.String(allow_none=True)

        @post_load
        def make(self, data):
            return Address(**data)


class Customer(Resource):
    """
    https://dev.chartmogul.com/v1.0/reference#customers
    """
    _path = "/customers{/uuid}"
    _root_key = 'entries'
    _many = namedtuple('Customers',
                       [_root_key, "has_more", "per_page", "page", "current_page", "total_pages"])

    class _Schema(Schema):
        # All operations
        data_source_uuid = fields.String()
        external_id = fields.String()
        name = fields.String()
        company = fields.String()
        email = fields.String()
        city = fields.String(allow_none=True)
        state = fields.String(allow_none=True)
        country = fields.String(allow_none=True)
        zip = fields.String(allow_none=True)
        lead_created_at = fields.DateTime(allow_none=True)
        free_trial_started_at = fields.DateTime(allow_none=True)

        # Things that differ between create/update & retrieve/list
        attributes = fields.Nested(Attributes._Schema)

        # Retrieve/List only
        id = fields.Int()
        uuid = fields.String()
        external_ids = fields.List(fields.String())
        data_source_uuids = fields.List(fields.String())
        status = fields.String()
        customer_since = fields.DateTime(load_from="customer-since", allow_none=True)
        mrr = fields.Number()
        arr = fields.Number()
        billing_system_url = fields.String(load_from="billing-system-url", allow_none=True)
        chartmogul_url = fields.String(load_from="chartmogul-url")
        billing_system_type = fields.String(load_from="billing-system-type")
        currency = fields.String()
        currency_sign = fields.String(load_from="currency-sign")
        address = fields.Nested(Address._Schema, allow_none=True)

        @post_load
        def make(self, data):
            return Customer(**data)

    _schema = _Schema(strict=True)


Customer.search = Customer._method('all', 'get', '/customers/search')
Customer.merge = Customer._method('merge', 'post', '/customers/merges')
Customer.connectSubscriptions = Customer._method('create', 'post', '/customers/{uuid}/connect_subscriptions')



from marshmallow import Schema, fields, post_load
from ..resource import Resource, DataObject

class Attributes(Resource):
    """
    https://dev.chartmogul.com/v1.0/reference#customer-attributes
    """
    _path = "/customers{/uuid}/attributes"

    class _Schema(Schema):
        tags = fields.List(fields.String())
        stripe = fields.Dict()
        clearbit = fields.Dict()
        custom = fields.Dict()

        @post_load
        def make(self, data):
            return Attributes(**data)

    _schema = _Schema(strict=True)
